var WIDTH=15
var HEIGHT=12
var FLOWTIMER=3

var STATE_SETUP=0
var STATE_GAMEPLAY=1
var STATE_GAMEOVER=2
var currentState = STATE_SETUP;

var score=0;

var UP = 0, RIGHT = 1, DOWN = 2, LEFT = 3;

var TILES = [
	[],
	[UP],
	[RIGHT],
	[DOWN],
	[LEFT],
	[UP, DOWN],
	[LEFT, RIGHT],
	[UP, RIGHT],
	[RIGHT, DOWN],
	[DOWN, LEFT],
	[LEFT, UP],
	[LEFT, RIGHT, DOWN],
	[UP, DOWN, LEFT],
	[LEFT, UP, RIGHT],
	[UP, RIGHT, DOWN],
	[UP, RIGHT, DOWN, LEFT]	
];

var $ctr;

//

var $ = function(sel) {
	if (sel[0] == "<") {
		return document.createElement(sel.slice(1, -1));
	} else {
		return document.querySelectorAll(sel);
	}
};

//

var opposite = function(direction) {
	return [DOWN, LEFT, UP, RIGHT][direction];
};

var log = function(message) {
	$log = $("#log")[0];
	$log.textContent += message + "\n";
	$log.scrollTop = $log.scrollHeight;
};

// 

var cell_get_coordinates = function($cell) {
	var XPOS = -1, YPOS = -1;
			
	for (var i = 0; i < WIDTH; ++i) {
		if ($cell.parentElement.children[i] == $cell) {
			XPOS = i;
			break;
		}
	}
	
	for (var i = 0; i < HEIGHT; ++i) {
		if ($ctr.children[i] == $cell.parentElement) {
			YPOS = i;
			break;
		}
	}
	
	return [XPOS, YPOS]; // 0,0 in the top-left corner
};

var cell_from_coordinates = function(x, y) {
	if (x >= WIDTH || x < 0 || y >= HEIGHT || y < 0) {
		return null;
	} else {
		return $ctr.children[y].children[x];
	}
};

var cell_get_tidx = function($cell) {
	var attr = ("" + $cell.style['background-position']);
	if (attr === "") {
		return 0;
	} else {
		return parseInt( attr.split(" ")[1], 10 ) / -40;
	}
};

var cell_neighbour = function($cell, direction) {
	var ctx = cell_get_coordinates($cell);
	switch(direction) {
		case UP:    return cell_from_coordinates(ctx[0]    , ctx[1] - 1);
		case RIGHT: return cell_from_coordinates(ctx[0] + 1, ctx[1]    );
		case DOWN:  return cell_from_coordinates(ctx[0]    , ctx[1] + 1);
		case LEFT:  return cell_from_coordinates(ctx[0] - 1, ctx[1]    );
	}
};

var cell_connected_neighbours = function($cell, directions) {
	var ret = [];
	for (var i = 0; i < directions.length; ++i) {
		var $neighbour = cell_neighbour($cell, directions[i]);
		if ($neighbour == null) {
			continue;
		}
		
		var entry_dirs = TILES[cell_get_tidx($neighbour)];
		
		if ($neighbour !== null) {
			if (entry_dirs.indexOf( opposite(directions[i]) ) !== -1 ) {
				ret.push($neighbour);
			}
		}
	}
	return ret;
};

// 

var boardSetup = function() {
	$ctr = $("#ctr")[0]; // assign to global
	
	for(var i = 0; i < HEIGHT; ++i) {
		var $row = $("<div>");
		$row.classList = "row";
		
		for(var j = 0; j < WIDTH; ++j) {
			var $cell = $("<div>");
			$cell.classList = "cell";
			$row.appendChild( $cell );
		}
		
		$ctr.appendChild($row);
	}
};

var handleClick = function($cell) {
	if (currentState == STATE_GAMEOVER || $cell.classList.contains("locked")) {
		return;
	}
	
	// pick new tile
	var tidx = Math.floor(Math.random() * TILES.length);
	$cell.style['background-position'] = "0 -"+(tidx * 40)+"px";
};

var clickSetup = function() {
	window.addEventListener("click", function(ev) {
		if (ev.target.classList.contains("cell")) {
			handleClick(ev.target);
			ev.stopPropagation();
		}
	});
};

var flowTargets = function() {
	var $locked = $(".cell.locked");	
	var $targets = [];
	
	for (var i = 0; i < $locked.length; ++i) {
		var exit_dirs = TILES[cell_get_tidx($locked[i])];
		var $neighbours = cell_connected_neighbours($locked[i], exit_dirs);
		
		// Lock neighbours only if they're connected
		for (var j = 0; j < $neighbours.length; ++j) {
			if (! $neighbours[j].classList.contains("locked")) {
				$targets.push($neighbours[j]);
			}
		}
	}
	
	return $targets;
};

var flowSetup = function() {
	setInterval(function() {
		if (currentState != STATE_GAMEPLAY) {
			return; // inactive game
		};
		
		var $locked = $(".cell.locked");
		
		var $targets = flowTargets();
		var did_flow = ($targets.length > 0);
		
		if (did_flow) {
			for (var i = 0; i < $targets.length; ++i) {
				$targets[i].classList.add("locked");
			}
			score++;
		} else {
			// gameover
			currentState = STATE_GAMEOVER;
			log("Game over! Score " + score);
			log("Press F5 to try again.");
		}
				
	}, FLOWTIMER*1000);
};

var startingPieceSetup = function() {
	var $cell = cell_from_coordinates(
		Math.floor(Math.random() * (WIDTH-2))+1,
		Math.floor(Math.random() * (HEIGHT-2))+1
	);
	
	// Assign tile 1/2/3/4
	var tidx = Math.floor(Math.random() * 4) + 1;
	$cell.style['background-position'] = "0 -"+(tidx * 40)+"px";
	
	// Lock
	$cell.classList.add("locked");
	
	setTimeout(function() {
		log("5 seconds left!");
	}, 5*1000);
	
	setTimeout(function() {
		log("Flow start!");
		currentState = STATE_GAMEPLAY;
	}, 10*1000);
};

var aiSetup = function() {
	setInterval(function() {
		if ($("#ai")[0].checked) {
			var best_candidates  = [];
			var other_candidates = [];
			
			// All squares
			var $all = $(".cell:not(.locked)");
			
			// For each square, find neighbours
			for (var i = 0; i < $all.length; ++i) {
				
				// If there are neighbours connecting into this square, it's a candidate
				var $inbound = cell_connected_neighbours($all[i], [UP, DOWN, LEFT, RIGHT]);				
				
				// If the neighbours are locked, it's a good candidate
				var is_good_candidate = false;
				for (var j = 0; j < $inbound.length; ++j) {
					if ($inbound[j].classList.contains("locked")) {
						best_candidates.push($all[i]);
						is_good_candidate = true;
						break;
					}
				}
				
				if ($inbound.length && !is_good_candidate) {
					other_candidates.push($all[i]);
				}
			}
			
			// Check if the flow is only going into endpieces
			var $targets = flowTargets();
			if ($targets.length > 0) {
				
				var endpiece_ct = 0;
				for (var i = 0; i < $targets.length; ++i) {
					var tidx = cell_get_tidx($targets[i]);
					if (tidx >= 1 && tidx <= 4) {
						endpiece_ct++;
					}
				}
				if (endpiece_ct == $targets.length) {
					// The only flow targets were endpieces!
					$targets = []; // Kick below code into using best candidates only
				}
			}
			
			// If the flow has somewhere to go, use other_candidates,
			//  otherwise use best_candidates.
			
			if (best_candidates.length && $targets.length == 0) {
				handleClick(best_candidates[  Math.floor(Math.random() * best_candidates.length ) ]);
			} else if (other_candidates.length) {
				handleClick(other_candidates[ Math.floor(Math.random() * other_candidates.length) ]);
			}
		}
	}, 0.5*1000); // 4 clicks per flow
};

window.addEventListener("load", function() {
	boardSetup();
	clickSetup();
	flowSetup();
	
	log("10 second headstart!");
	startingPieceSetup();
	
	aiSetup();
});
